class NoColumnsError(Exception):
    """
    Raised if no header are specified
    """
    pass


class NoTableNameGivenError(ValueError):
    """
    Raised if no header are specified
    """
    pass



from sqlalchemy.sql.schema import Table
from sqlalchemy.sql import exists, and_, select, func, or_, insert
from sqlalchemy.sql.functions import now
from sqlalchemy.sql.expression import literal_column
from datetime import datetime

from ninjasql.db.sqa_table_loads import TableLoad


class SqaExtractor(object):
    """
    :param staging_table: Sqa Table class object
    :param history_table: Sqa Table class object
    :param logical_pk: Logical primary key of the target table as list
    :param con: Sqlalchemy database connection
    :param load_strategy: [jinja, database_table]
    """

    def __init__(self,
                 staging_table: Table,
                 history_table: Table,
                 logical_pk: list,
                 con,
                 load_strategy: str,
                 ):
        self._staging_table = staging_table
        self._history_table = history_table
        self._logical_pk = logical_pk
        self._con = con
        self._load_strategy = load_strategy

        ALLOWED_STRATEGIES = ['jinja', 'database_table']

        if not isinstance(self._staging_table, Table):
            raise TypeError("Must be a SQA Table class instance")
        if self._load_strategy not in ALLOWED_STRATEGIES:
            stra = ' ,'.join(ALLOWED_STRATEGIES)
            raise ValueError(f"Invalid load stragey. Allowed are: '{stra}'")

    def get_col_names(self) -> list:
        """
        method that returns all columns
        COLUMN class has:
        {'key': X, 'name': Y, 'table': T}
        """
        return [c.name for c in self._staging_table.c]

    def get_his_col_names(self) -> list:
        """
        method that returns all columns
        COLUMN class has:
        {'key': X, 'name': Y, 'table': T}
        """
        return [c.name for c in self._history_table.c]

    def get_source_col_names(self) -> list:
        """
        method returns all non technical or business related column names
        """
        return [i for i, j in zip(
            self.get_his_col_names(),
            self.get_col_names()) if i == j]

    def get_table_name(self) -> str:
        """
        method that returns all columns
        COLUMN class has:
        {'key': X, 'name': Y, 'table': T}
        """
        return [c.table.name for c in self._staging_table.c][0]

    def get_hist_table_name(self) -> str:
        """
        method that returns all columns
        COLUMN class has:
        {'key': X, 'name': Y, 'table': T}
        """
        return [c.table.name for c in self._history_table.c][0]

    def get_staging_columns(self) -> list:
        """
        method that returns all staging column
        """
        return [getattr(self._staging_table.c, c) for
                c in self.get_col_names()]

    def set_batch_date(self) -> str:
        """
        Set the correct batch date either table load
        or jinja expression
        """
        if self._load_strategy == "jinja":
            return select([func.date(
                literal_column(
                    r"{{ batch_date }}")).label("BATCH_RUN_AT")]).as_scalar()
        elif self._load_strategy == "database_table":
            table = self.get_hist_table_name()
            return select([TableLoad.BatchDate]).where(
                TableLoad.name == table).as_scalar()

    def set_validto_date(self) -> str:
        """
        Set the correct batch date either table load
        or jinja expression
        """
        if self._load_strategy == "jinja":
            return select([func.date(
                literal_column(
                    r"{{ validto_date }}")).label(
                        "VALID_TO_DATE")]).as_scalar()
        elif self._load_strategy == "database_table":
            table = self.get_hist_table_name()
            return select([TableLoad.ValidToDate]).where(
                TableLoad.name == table).as_scalar()

    def set_subquery_validto_date(self) -> str:
        """
        Set the correct validto date either table load
        or jinja expression in a subquery
        """
        if self._load_strategy == "jinja":
            return func.date(literal_column(r"{{ validto_date }}"))
        elif self._load_strategy == "database_table":
            return select([TableLoad.ValidToDate]).where(
                TableLoad.name == self.get_hist_table_name())

    def set_subquery_batch_date(self) -> str:
        """
        Set the correct batch date either table load
        or jinja expression in a subquery
        """
        if self._load_strategy == "jinja":
            return func.date(literal_column(r"{{ batch_date }}"))
        elif self._load_strategy == "database_table":
            return select([TableLoad.BatchDate]).where(
                TableLoad.name == self.get_hist_table_name())

    def set_subquery_offsetvalidto_date(self) -> str:
        """
        Set the correct offset validto date either table load
        or jinja expression in a subquery
        """
        if self._load_strategy == "jinja":
            return func.date(literal_column(r"{{ offset_validto_date }}"))
        elif self._load_strategy == "database_table":
            return select([TableLoad.OffsetValidToDate]).where(
                TableLoad.name == self.get_hist_table_name())

    def set_validfrom_date(self) -> str:
        """
        Set the correct batch date either table load
        or jinja expression
        """
        if self._load_strategy == "jinja":
            return select([func.date(
                literal_column(
                    r"{{ validfrom_date }}")).label(
                        "VALID_FROM_DATE")]).as_scalar()
        elif self._load_strategy == "database_table":
            table = self.get_hist_table_name()
            return select([TableLoad.ValidFromDate]).where(
                TableLoad.name == table).as_scalar()

    def set_metadata_colums(self,
                            batch_dt: str = None,
                            from_dt: str = None,
                            to_dt: str = None) -> list:
        """
        method that create the metadata columns for the insert
        and updates as scalar selectable
        """
        valid_to_dt = self.set_validto_date()  # "'9999-12-31'"
        valid_from_dt = self.set_validfrom_date()  # "'2020-01-31'"
        batch_dt = self.set_batch_date()  # "'2020-02-01'"

        metadata_col = []
        metadata_col.append(
            select([now().label("UPDATED_AT")]).as_scalar()
        )
        metadata_col.append(batch_dt)
        metadata_col.append(valid_from_dt)
        metadata_col.append(valid_to_dt)
        return metadata_col

    def def_equal_pk_col(self) -> list:
        """
        method that returns all staging column
        """
        pk_cols = []
        for pk_col in self._logical_pk:
            pk_cols.append(
                getattr(self._staging_table.c, pk_col) ==
                getattr(self._history_table.c, pk_col)
            )
        return pk_cols

    def get_compare_columns(self) -> list:
        """
        method that returns all compare columns
        """
        compare_columns = []
        for col in self.get_source_col_names():
            compare_columns.append(
                getattr(self._staging_table.c, col) !=
                getattr(self._history_table.c, col)
            )
        return compare_columns

    def get_staging_table_pk_col(self) -> list:
        """
        method that gets all pk columns from staging table
        """
        return [getattr(self._staging_table.c, pk) for pk in self._logical_pk]

    def get_tableload_insert(self) -> str:
        """
        method that generate the default INSERT of the persistent staging table
        for the table load table if strategy == database_table
        """
        valid_to_dt = datetime(9999, 12, 31)  # "'9999-12-31'"
        valid_from_dt = datetime(2020, 1, 31)  # "'2020-01-31'"
        batch_dt = datetime(2020, 2, 1)  # "'2020-02-01'"
        offsetvalid_to_dt = datetime(2020, 1, 30)
        ins = insert(TableLoad).values(
            name=self.get_hist_table_name(),
            BatchDate=batch_dt,
            ValidToDate=valid_to_dt,
            ValidFromDate=valid_from_dt,
            OffsetValidToDate=offsetvalid_to_dt)

        return str(ins.compile(bind=self._con,
                               compile_kwargs={"literal_binds": True}))

    def scd2_new_insert(self) -> str:
        """
        method that insert statement
        """
        all_his_columns = self.get_his_col_names()
        all_stg_columns = self.get_staging_columns()
        all_stg_columns.extend(self.set_metadata_colums())
        exist_stat = self.get_staging_table_pk_col()
        filters = self.def_equal_pk_col()

        stmt = (self._history_table.insert().
                from_select(all_his_columns,
                select(all_stg_columns).where(
                    ~exists(exist_stat).where(and_(
                        *filters)))
                        ))
        return str(stmt.compile(bind=self._con,
                                compile_kwargs={"literal_binds": True}))

    def scd2_updated_insert(self) -> str:
        """
        scd2 command for all records where an update is needed due to a
        change record in a source table. Insert a new record for an existing
        but changed dataset
        """
        # compare columns
        compare_columns = self.get_compare_columns()
        # all columns
        all_stg_columns = self.get_staging_columns()
        # get join columns
        pk_cols = self.def_equal_pk_col()

        all_stg_columns.extend(self.set_metadata_colums())

        sel = (select(all_stg_columns).
               select_from(self._history_table.join(
                   self._staging_table,
                   and_(*pk_cols))).where(
                       or_(*compare_columns)).where(
                           self._history_table.c.VALID_TO_DATE
                           == self.set_subquery_validto_date()))

        stmt = (self._history_table.insert().
                from_select(self.get_his_col_names(), sel))
        return str(stmt.compile(bind=self._con,
                                compile_kwargs={"literal_binds": True}))

    def scd2_updated_update(self) -> str:
        """
        scd2 command for all records where an update is needed due to a change
        in the source system. Set current latest record VALID_TO_DATE to
        current batch date
        """
        # compare columns
        compare_columns = self.get_compare_columns()
        exist_stat = self.get_staging_table_pk_col()
        filters = self.def_equal_pk_col()

        to_dt = self.set_subquery_validto_date()  # "'9999-12-31'"
        batch_dt = self.set_subquery_batch_date()  # "'2020-02-01'"
        offset_validto = self.set_subquery_offsetvalidto_date()
        # from_dt = "'2020-01-31'"

        upd = (self._history_table.update()
               .values(
                    VALID_TO_DATE=offset_validto,
                    UPDATED_AT=now()).where(
                        exists(exist_stat).where(
                            and_(*filters)).where(
                                or_(*compare_columns)).where(
                                    and_(self._history_table.c.VALID_TO_DATE
                                         == to_dt,
                                         self._history_table.c.BATCH_RUN_AT
                                         < batch_dt
                                         ))))
        return str(upd.compile(bind=self._con,
                               compile_kwargs={"literal_binds": True}))

    def scd2_deleted_update(self) -> str:
        """
        scd2 command for all records where an update is needed due to a change
        in the source system. Set current latest record VALID_TO_DATE to
        current batch date
        """
        to_dt = self.set_subquery_validto_date()  # "'9999-12-31'"
        # batch_dt = self.set_subquery_batch_date()  # "'2020-02-01'"
        offset_validto = self.set_subquery_offsetvalidto_date()

        no_exist = self.get_staging_table_pk_col()
        filters = self.def_equal_pk_col()

        upd = (self._history_table.update()
               .values(
                    VALID_TO_DATE=offset_validto,
                    UPDATED_AT=now()).where(
                        and_(~exists(no_exist).where(
                            and_(*filters)),
                             self._history_table.c.VALID_TO_DATE
                             == to_dt)))
        return str(upd.compile(bind=self._con,
                               compile_kwargs={"literal_binds": True}))



from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import Column, String, DateTime
from sqlalchemy.schema import CreateTable

Base = declarative_base()


class TableLoad(Base):
    __tablename__ = 'tableloads'

    name = Column(String, primary_key=True)
    BatchDate = Column(DateTime)
    ValidToDate = Column(DateTime)
    OffsetValidToDate = Column(DateTime)
    ValidFromDate = Column(DateTime)


def get_sqa_tableload(con):
    """
    param con: sqa database engine
    """
    return str(CreateTable(TableLoad.__table__).compile(con))



import networkx as nx
from typing import Iterator


class Singleton:

    def __init__(self, cls):
        self._cls = cls

    def Instance(self):
        try:
            return self._instance
        except AttributeError:
            self._instance = self._cls()
            return self._instance

    def __call__(self):
        raise TypeError('Singletons must be accessed through `Instance()`.')

    def __instancecheck__(self, inst):
        return isinstance(inst, self._cls)


@Singleton
class TableDep(object):

    def __init__(self):
        self.graph = nx.DiGraph()

    def addTable(self, name: str, depends_on: str) -> None:
        self.graph.add_edge(name, depends_on)

    def show_edges(self) -> list:
        """
        method that list all edges of the graph
        """
        return list(self.graph.edges())

    def find_path(self) -> list:
        """
        method that returns path in topological Sorting
        """
        return list(reversed(list(nx.topological_sort(self.graph))))

    def find_batch_path(self) -> list:
        """
        method that returns the topological Sorting with a batch
        grouping of all tasks that can run at the same time
        """
        return list(self._create_batch_order())

    def _create_batch_order(self) -> Iterator[list]:
        """
        method that create batches of nodes in order to run concurrent or
        parallel task loads
        """
        inverse_graph = nx.DiGraph()
        for name, depends_on in self.show_edges():
            inverse_graph.add_edge(depends_on, name)
        indegree_map = {v: d for v, d in inverse_graph.in_degree() if d > 0}
        zero_indegree = [v for v, d in inverse_graph.in_degree() if d == 0]
        while zero_indegree:
            yield zero_indegree
            new_zero_indegree = []
            for v in zero_indegree:
                for _, child in inverse_graph.edges(v):
                    indegree_map[child] -= 1
                    if not indegree_map[child]:
                        new_zero_indegree.append(child)
            zero_indegree = new_zero_indegree



from pathlib import Path
import logging
import configparser

logging.basicConfig(level=logging.INFO,
                    format='[%(asctime)s %(name)s %(levelname)s:%(message)s]')
log = logging.getLogger(__name__)


class Config:
    """
    :ini_path : path of the ini file for the configuration
    """
    SECTIONS = {
        "Staging": [
            "schema_name",
            "table_prefix_name"
        ],
        "PersistentStaging": [
            "schema_name",
            "table_prefix_name"
        ]
    }

    def __init__(self,
                 ini_path: str = ''):
        self._ini_path = Path(ini_path)
        self._config = None

    @property
    def ini_path(self):
        """
        Get ini path
        """
        return self._ini_path

    @ini_path.setter
    def ini_path(self, value):
        """
        change ini path
        """
        try:
            self._ini_path = Path(value)
        except Exception as e:
            log.error(f"Please provide a valid path. Error: {e}")

    @property
    def config(self):
        """
        Get config
        """
        return self._config

    def read(self) -> None:
        """
        method that reads a ini file and its content
        """
        self._config = configparser.ConfigParser()
        self._config.read(self._ini_path)
        self._check_content()

    def _check_content(self) -> None:
        """
        method that reads expected sections and options of the
        ini file
        """
        for key, value in self.__class__.SECTIONS.items():
            for sec in value:
                try:
                    self._config.get(key, sec)
                except configparser.NoSectionError as e:
                    log.error(f"Can't find Section: {key} in ini file."
                              "Error: {e}")
                    raise e
                except configparser.NoOptionError as e:
                    log.error(f"Can't find option: {sec} in section {key} "
                              "ini file. Error: {e}")
                    raise e



from pathlib import Path
import os
import logging
import pandas as pd
from pandas import DataFrame
from pandas.io.sql import (
    pandasSQL_builder,
    SQLTable,
    SQLDatabase,
    SQLiteTable)
from sqlalchemy.schema import CreateSchema
from sqlalchemy import inspect
from datetime import datetime
import traceback

from ninjasql.errors import NoColumnsError, NoTableNameGivenError
from ninjasql.settings import Config
from ninjasql.db.sqa_dml_extractor import SqaExtractor
from ninjasql.db.sqa_table_loads import get_sqa_tableload
from ninjasql.dep.table_dependency import TableDep

logging.basicConfig(level=logging.INFO,
                    format='[%(asctime)s %(name)s %(levelname)s:%(message)s]')
log = logging.getLogger(__name__)


class FileInspector(object):
    """
    1. There is a source file with data in it
    2. Profile data, get metadata
    3. Create a the needed database tables
        a. staging table
        b. historization table
    4. There need to be an import folder
    5. Archive folder
    :param file: filepath
    :param seperator: file seperator for csv and txt files
    :param header: Is a header row with columns names given
    :param type: File type {csv, json}
    :param columns: Custom column names if no header is given
    :param orient: Json orientation
    :param con: Sqlalchemy database connection
    """
    def __init__(self,
                 cfg_path: str,
                 file: str = None,
                 seperator: str = ',',
                 header: int = 0,
                 type: str = None,
                 columns: list = None,
                 orient: str = 'records',
                 con=None
                 ):
        self._cfg_path = cfg_path
        self._file = file
        self._seperator = seperator
        self._header = header
        self._type = type
        self._columns = columns
        self._orient = orient
        self._data = None
        self._his_data = None
        self._con = con
        self.config = Config()
        self._Dag = TableDep.Instance()

        self.load_config(cfg_path=self._cfg_path)

    def _has_header(self) -> bool:
        """
        Instance method that checks if header exist or not
        """
        if self._header is None and self._columns is None:
            return False
        return True

    @property
    def file(self):
        """
        Get file
        """
        return self._file

    @file.setter
    def file(self, value):
        """
        change file
        """
        try:
            self._file = Path(value)
        except Exception as e:
            log.error(f"Please provide a valid path. Error: {e}")

    def load_config(self, cfg_path: str) -> None:
        """
        method that load the ini configuration file
        """
        self.config.ini_path = cfg_path
        self.config.read()

    def _read_data(self) -> None:
        """
        Method that reads data and save it as a instance variable
        """
        if not self._is_file():
            log.error(f"Can't find the a file. Check file path!")
        if self._type == 'csv':
            self._csv_reader()
        elif self._type == 'json':
            self._json_reader()

    def _build_header(self) -> None:
        """
        Instance method that implements a correct header
        TODO: This method has side effects
        """
        if not self._has_header():
            self._columns = None

    def _csv_reader(self) -> None:
        """
        Instance method that implements a csv reader
        TODO: This method has side effects
        """
        self._build_header()
        try:
            self._data = pd.read_csv(
                filepath_or_buffer=self._file,
                sep=self._seperator,
                header=self._header,
                names=self._columns)
        except Exception:
            track = traceback.format_exc()
            log.error(f"Upps. Check file and location. Error: {track}")

    def _json_reader(self):
        """
        Instance method that implements a json reader
        """
        self._build_header()
        try:
            self._data = pd.read_json(
                path_or_buf=self._file,
                orient=self._orient)
        except Exception:
            track = traceback.format_exc()
            log.error(f"Upps. Check file and location. Error: {track}")

    def show_columns(self) -> list:
        """
        Method that shows all columns of a provided dataset
        """
        self._load_df_if_empty()
        return self._data.columns

    def get_dtypes(self) -> dict:
        """
        Method that get columns datatype as dict
        """
        self._load_df_if_empty()
        if not self._has_header():
            log.error("Column datatypes are asked but no columns are "
                      "specified or given in a file.")
            raise NoColumnsError
        return self._data.dtypes.to_dict()

    def col_to_str(self) -> None:
        """
        Method that change all column data type to a string type
        """
        self._load_df_if_empty()
        self._data = self._data.astype(str)

    def _is_file(self) -> bool:
        """
        Checks if a given file exist or not
        """
        self._file = Path(self._file)
        return self._file.is_file()

    def _load_df_if_empty(self) -> None:
        """
        Instance method that loads data in a pandas df
        if not already done.
        TODO: This method has side effects
        """
        if self._data is None:
            self._read_data()

    def _set_path(self, path: str) -> Path:
        """
        Instance method that converts a str to a Path object
        """
        try:
            tpath = Path(path)
            return tpath
        except Exception as e:
            log.error(f"Please provide a valid path. Error: {e}")
            raise e

    def save_staging_ddl(self,
                         path,
                         table_name: str,
                         schema: str = None,
                         database: str = None,
                         dtype=None) -> None:
        """
        Method that get the sql ddl statement and save it as a file
        in a target path
        :param path: Directory path where file should be saved
        :table name: DDL table name
        :schema: DDL schema name. If not specified take schema from the
        configuration ini file
        :database: DDL database name
        :dtype : dict of column name to SQL type, default None
        Optional specifying the datatype for columns. The SQL type should
        be a SQLAlchemy type, or a string for sqlite3 fallback connection.
        """
        self._load_df_if_empty()
        tpath = self._set_path(path=path)

        try:
            qu_name = self._build_name(
                table=table_name,
                db=database,
                schema=schema,
                table_type="staging"
            )
            ddl = self._extract_ddl(
                frame=self._data,
                name=qu_name,
                dtype=dtype
            )
            self._save_file(
                path=tpath,
                fname=qu_name,
                content=ddl,
                subdir='DDL'
            )
        except ValueError as e:
            log.error(f"{e}")

    def _extract_ddl(self,
                     frame: DataFrame,
                     name: str,
                     dtype: dict) -> str:
        """
        Instance method that extracts the ddl statement of the
        pandals sql core module
        """
        return pd.io.sql.get_schema(frame=frame,
                                    name=name,
                                    con=self._con,
                                    dtype=dtype)

    def save_history_ddl(self,
                         path,
                         table_name: str,
                         schema: str = None,
                         database: str = None,
                         dtype=None) -> None:
        """
        Method that get the sql history ddl statement and save it as a file
        in a target path
        :param path: Directory path where file should be saved
        :table name: DDL table name
        :schema: DDL schema name
        :database: DDL database name
        :dtype : dict of column name to SQL type, default None
        Optional specifying the datatype for columns. The SQL type should
        be a SQLAlchemy type, or a string for sqlite3 fallback connection.
        """
        self._load_df_if_empty()
        tpath = self._set_path(path=path)

        self._add_scd2_attributes()
        try:
            qu_name = self._build_name(table=table_name,
                                       db=database,
                                       schema=schema,
                                       table_type="history")
            ddl = self._extract_ddl(
                frame=self._his_data,
                name=qu_name,
                dtype=dtype
            )

            self._save_file(
                path=tpath,
                fname=qu_name,
                content=ddl,
                subdir='DDL'
            )
        except ValueError as e:
            log.error(f"{e}")

    def _add_scd2_attributes(self) -> None:
        """
        Instance method that add scd2 relevant attributes
        """
        self._load_df_if_empty()
        now = datetime.now()
        now_str = now.strftime("%Y-%m-%d %H:%M:%S")
        self._his_data = self._data.copy()
        self._his_data['UPDATED_AT'] = pd.Timestamp(now_str)
        self._his_data['BATCH_RUN_AT'] = pd.Timestamp(now_str)
        self._his_data['VALID_FROM_DATE'] = pd.Timestamp(now_str)
        self._his_data['VALID_TO_DATE'] = pd.Timestamp(now_str)

    def _build_name(self,
                    table: str,
                    table_type: str,
                    db: str = None,
                    schema: str = None
                    ) -> str:
        """
        Method that create the final qualified sql object name
        """
        if table_type == 'staging':
            t_pre = self.config.config['Staging']['table_prefix_name']
            schema = schema or self.config.config['Staging']['schema_name']
        elif table_type == 'history':
            sec = 'PersistentStaging'
            t_pre = self.config.config[sec]['table_prefix_name']
            schema = schema or self.config.config[sec]['schema_name']

        if table is None:
            log.error(f"No table name given but needed. Please specify name:")
            raise NoTableNameGivenError
        if (db and schema):
            return f"{db}.{schema}.{t_pre}_{table}"
        elif (db and not schema):
            return f"{db}.{t_pre}_{table}"
        elif (schema and not db):
            return f"{schema}.{t_pre}_{table}"
        elif (table and not schema and not db):
            return f"{t_pre}_{table}"

    def _save_file(self,
                   path: str,
                   fname: str,
                   content: str,
                   subdir: str = None,
                   ) -> None:
        """
        method that save a string content to a given path.
        :param path: Directory path where file shall be saved
        :fname : Target filename
        :content: File content
        :subdir: If a subdir is given the file is saved in the subdir
        """
        if not os.path.isdir(path):
            log.error(f"Given Path is not a valid directory. Please check!")
            raise FileNotFoundError

        basename = f"{fname.replace('.', '_')}"
        modelname = basename.split('_')[-1]
        nfname = f"{basename}.sql"

        if subdir:
            modelname = os.path.join(modelname, subdir)

        Path(os.path.join(path, modelname)).mkdir(parents=True, exist_ok=True)

        with open(os.path.join(path, modelname, nfname), "w") as f:
            f.write(content)

    def create_db_table(self,
                        table_name: str,
                        type: str,
                        schema: str = None,
                        if_exists: str = 'replace',
                        dtype=None
                        ) -> None:
        """
        method that creates the database table without data
        :table_name path: target table name
        :schema : Target database schema name
        :if_exists: How to behave if the table already exists.
        {‘fail’, ‘replace’, ‘append’}
        :dtype : dict of column name to SQL type, default None
        Optional specifying the datatype for columns. The SQL type should
        be a SQLAlchemy type, or a string for sqlite3 fallback connection.
        :type : str {'staging', 'history'}. Determine if the staging or
        history table shall be created
        """
        self._load_df_if_empty()
        if schema:
            insp = inspect(self._con)
            schemas = insp.get_schema_names()
            if schema not in schemas:
                self._con.execute(CreateSchema(schema))
        try:
            if type == "staging":
                df = self._data[:0]
            elif type == "history":
                df = self._his_data[:0]
            df.to_sql(
                name=table_name,
                schema=schema,
                con=self._con,
                if_exists=if_exists,
                index=False,
                dtype=dtype
            )
        except Exception as e:
            log.error(f"Can't create db table. Error: {e}")
            raise e

    def create_file_elt_blueprint(self,
                                  path: str,
                                  table_name: str,
                                  logical_pk: list,
                                  load_strategy: str,
                                  ):
        self._load_df_if_empty()
        self.save_staging_ddl(path=path, table_name=table_name)
        self.save_history_ddl(path=path, table_name=table_name)
        stg = self._get_sqa_table(
            table_name=table_name,
            table_type="staging")
        his = self._get_sqa_table(
            table_name=table_name,
            table_type="history")
        c = SqaExtractor(
            staging_table=stg,
            history_table=his,
            logical_pk=logical_pk,
            load_strategy=load_strategy,
            con=self._con)

        base_name = c.get_hist_table_name()
        self._save_file(
                path=path,
                fname=f"scd2_1_{base_name}",
                content=c.scd2_new_insert(),
                subdir='DML'
        )

        self._save_file(
                path=path,
                fname=f"scd2_2_{base_name}",
                content=c.scd2_updated_insert(),
                subdir='DML'
        )
        self._Dag.addTable(f"scd2_2_{base_name}.sql",
                           f"scd2_1_{base_name}.sql")

        self._save_file(
                path=path,
                fname=f"scd2_3_{base_name}",
                content=c.scd2_updated_update(),
                subdir='DML'
        )

        self._Dag.addTable(f"scd2_3_{base_name}.sql",
                           f"scd2_2_{base_name}.sql")

        self._save_file(
                path=path,
                fname=f"scd2_4_{base_name}",
                content=c.scd2_deleted_update(),
                subdir='DML'
        )
        self._Dag.addTable(f"scd2_4_{base_name}.sql",
                           f"scd2_3_{base_name}.sql")
        if load_strategy == 'database_table':
            self._save_file(
                    path=path,
                    fname="JOBTABLE_TABLELOAD",
                    content=get_sqa_tableload(con=self._con),
                    subdir='DDL'
            )

            self._save_file(
                path=path,
                fname=f"{base_name}_INSERT_TABLELOAD",
                content=c.get_tableload_insert(),
                subdir='DML'
            )

    def _get_sqa_table(self,
                       table_name: str,
                       table_type: str,
                       schema: str = None,
                       database: str = None,
                       dtype=None
                       ) -> None:
        """
        Method that extracts sqa table object
        """
        self._load_df_if_empty()
        pandas_sql = pandasSQL_builder(con=self._con)

        sqllite = False
        if not isinstance(pandas_sql, SQLDatabase):
            sqllite = True

        pd_db = SQLDatabase(engine=self._con)
        table_klass = SQLTable if not sqllite else SQLiteTable
        target_frame = (self._data if table_type == "staging"
                        else self._his_data)
        target_name = self._build_name(table=table_name, table_type=table_type)
        table = table_klass(
            name=target_name,
            pandas_sql_engine=pd_db,
            index=False,
            frame=target_frame
        )
        return table._create_table_setup()


if __name__ == "__main__":
    pass



import pathlib
from setuptools import setup, find_packages

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "readme.md").read_text()

# This call to setup() does all the work
setup(
    name="ninjasql",
    version="0.0.1",
    description="Create and extract SQL DDL, DMLs and scd2 blueprints.",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/dondaum/ninjasql",
    author="Sebastian Daum",
    author_email="sebastian.daum89@gmail.com",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",

        'Operating System :: Microsoft :: Windows',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: POSIX :: Linux',

        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    packages=find_packages(exclude=["tests", "tests.*"]),
    include_package_data=True,
    install_requires=[
        'decorator==4.4.2',
        'networkx==2.4',
        'numpy==1.18.2',
        'pandas==1.0.3',
        'python-dateutil==2.8.1',
        'pytz==2019.3',
        'six==1.14.0',
        'sqlalchemy==1.3.15'
    ],
    zip_safe=False,
)
